#include <Drivers.h>
#include <KernelExport.h>
#include <OS.h>
#include <module.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>

/*

	/dev/misc/lsmodules driver for BeOS
	(c) 2002, mmu_man, revol@free.fr
	lists loaded kernel modules
	usage:
		make driverinstall
		rescan lsmodules
		cat /dev/misc/lsmodules
	licenced under MIT licence
*/

/*
	NOTES:
	get_next_loaded_module_name()
	the cookie param is set to 0, then it gets the module_id of the returned module name.
	the id seems to be a unique number, starting to 1 and incrememted on each load of any module.
	the list is returned bottom-up (last loaded modules, highest module_id first)
*/


const char * device_names[]={"misc/lsmodules", NULL};
extern device_hooks lsmodules_hooks;

#define SHOW_HEADER
#define LSMODULES_HEADER "Loaded modules:\n"

typedef struct {
	uint32 id; /* current name */
	uint32 off; /* offset in the name */
	uint32 len; /* len of the buffer */
	uint32 count; /* not used */
	uint32 modid; /* module id (cookie to get_next_loaded_module()) */
	bool endoflist;
	char buffer[MAXPATHLEN+20];
} cookie_t;



status_t init_hardware(void) {
	return B_OK;
}

status_t init_driver(void) {
	return B_OK;
}

void uninit_driver(void) {
}

const char **publish_devices() {
	return device_names;
}

device_hooks *find_device(const char *name) {
	(void)name;
	return &lsmodules_hooks;
}

status_t lsmodules_open(const char *name, uint32 flags, cookie_t **cookie) {
	(void)name; (void)flags;
	*cookie = (void*)malloc(sizeof(cookie_t));
	if (*cookie == NULL) {
		dprintf("lsmodules_open : error allocating cookie\n");
		goto err0;
	}
	memset(*cookie, 0, sizeof(cookie_t));
	(*cookie)->endoflist = false;
#ifdef SHOW_HEADER
	strcpy((*cookie)->buffer, LSMODULES_HEADER);
	(*cookie)->len = strlen((*cookie)->buffer);
#endif
	return B_OK;
	
err1:
	dprintf("lsmodules_open : cleanup : will free cookie\n");
	free(*cookie);
	*cookie = NULL;
err0:
	return B_ERROR;
}

status_t lsmodules_close(void *cookie) {
	(void)cookie;
	return B_OK;
}

status_t lsmodules_free(cookie_t *cookie) {
	free(cookie);
	return B_OK;
}

status_t lsmodules_ioctl(cookie_t *cookie, uint32 op, void *data, size_t len) {
	(void)cookie; (void)op; (void)data; (void)len;
	return B_ERROR;
}

status_t lsmodules_read(cookie_t *cookie, off_t position, void *data, size_t *numbytes) {
	char buff[MAXPATHLEN];
	size_t len = MAXPATHLEN;
	(void)position;
	//dprintf("lsmodules_read : id = %ld, count = %ld, off = %ld\n", cookie->id, cookie->count, cookie->off);

	if (cookie->endoflist)
		goto ERROR_EOF;
	if (cookie->off >= cookie->len) {
		len = MAXPATHLEN;
		if (get_next_loaded_module_name(&(cookie->modid), buff, &len) < B_OK)
			goto ERROR_EOF;
		cookie->id++;
		sprintf(cookie->buffer, "%ld: %s\n", cookie->modid, buff);
		cookie->len = strlen(cookie->buffer);
		cookie->off = 0;
		cookie->id++;
	}
	//dprintf("lsmodules_read : name=%s\n", buff);

	//MIN()
	*numbytes = (((cookie->len) - (cookie->off)) < (*numbytes))?((cookie->len) - (cookie->off)):(*numbytes);
	memcpy(data, (cookie->buffer + cookie->off), *numbytes);
	cookie->off += *numbytes;
	return B_OK;
ERROR_EOF:
	cookie->endoflist = true;
	*numbytes = 0;
	return B_OK;
}

status_t lsmodules_write(void *cookie, off_t position, const void *data, size_t *numbytes) {
	(void)cookie; (void)position; (void)data;
	*numbytes = 0;
	return B_OK;
}

device_hooks lsmodules_hooks={
	(device_open_hook)lsmodules_open,
	lsmodules_close,
	(device_free_hook)lsmodules_free,
	(device_control_hook)lsmodules_ioctl,
	(device_read_hook)lsmodules_read,
	lsmodules_write,
	NULL,
	NULL,
	NULL,
	NULL
};
